/* This file is part of the Imote2 Codeloader.
 *
 * The Imote2 Codeloader is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * The Imote2 Codeloader is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with the Imote2 Codeloader. If not, see <http://www.gnu.org/licenses/>.
*/
/*
 * Configuration.java
 *
 * @author Markus Engel
 * Copyright (c) 2009,2010 Networked Systems Group, University of Kaiserslautern
 */
package imote;

import java.io.*;
import java.lang.reflect.*;
import java.util.*;
import org.apache.commons.cli.*;

/**
 * Configuration class
 */
public class Configuration {
	/* defaults */
	/** fallback vendor id if none given */
	public static final int DEFAULT_VENDOR_ID = 0x042b;
	/** fallback product id if none given */
	public static final int DEFAULT_PRODUCT_ID = 0x1337;

	/** represent an UploadRequest (more can be given) */
	public static class UploadRequest {
		/** default target address (application) */
		public static final long DEFAULT_TARGET_ADDRESS = 0x00400000l;
		/** default target address (configuration) */
		public static final long DEFAULT_CFG_ADDRESS = 0x00040000l;
		/** default target address (execute in RAM) */
		public static final long RAM_TARGET_ADDRESS = 0xffffffffl;
		/** file to be uploaded */
		private String fileName;
		/** target address */
		private long targetAddress = DEFAULT_TARGET_ADDRESS;
		/** execute in ram */
		private boolean executeRam = false;

		/** parse a target address and set it
		 *
		 * @param targetAddress address to be parsed
		 */
		public void setTargetAddress(String targetAddress) {
			long addr;
			if (targetAddress.equalsIgnoreCase("cfg")) {
				addr = DEFAULT_CFG_ADDRESS;
			} else {
				addr = Long.parseLong(cutRadix(targetAddress), getRadix(
					targetAddress));
				if (addr > 0xffffffffl) {
					/* 32 bit unsigned integer */
					throw new NumberFormatException();
				}
			}
			this.targetAddress = addr;
		}

		/** getter for the target address
		 *
		 * @return target address
		 */
		public long getTargetAddress() {
			return this.targetAddress;
		}

		/** write the target address to a stream such that it can be parsed
		 * by setTargetAddress again
		 *
		 * @param level indentaion level
		 * @param writer stream to be written to
		 */
		void printTargetAddress(int level, PrintStream writer) {
			writer.printf("0x%08x", this.targetAddress);
		}

		/** parse filename and set it
		 *
		 * @param fileName path of file to be used
		 */
		public void setFileName(String fileName) {
			this.fileName = fileName;
		}

		/** getter for filename
		 *
		 * @return filename
		 */
		public String getFileName() {
			return this.fileName;
		}

		/** write the filename to a stream such that it can be parsed by
		 * setFileName again
		 *
		 * @param level indentaion level
		 * @param writer stream to be written to
		 */
		void printFileName(int level, PrintStream writer) {
			writer.print(this.fileName);
		}

		/** getter for the execute-in-ram flag
		 *
		 * @return execute-in-ram flag
		 */
		public boolean getExecuteRam() {
			return this.executeRam;
		}

		/** simple test to determine wether the given file is to be executed
		 *
		 * @return true iff the file in that request is about to be executed
		 */
		public boolean isExecutable() {
			return this.executeRam
				|| this.targetAddress == DEFAULT_TARGET_ADDRESS;
		}

		/** setter for the execute-in-ram flag
		 *
		 * @param execMem wether that request should be executed in ram
		 */
		public void setExecuteRam(boolean execMem) {
			this.executeRam = execMem;
		}

		/** write the execute-in-ram flag to a stream
		 *
		 * @param level indentaion level
		 * @param writer stream to be written to
		 */
		void printExecuteRam(int level, PrintStream writer) {
			writer.print(new Boolean(this.executeRam).toString());
		}
	}

	/* configuration and defaults */
	/** USB vendor id */
	private int vendorId = DEFAULT_VENDOR_ID;
	/** USB product id */
	private int productId = DEFAULT_PRODUCT_ID;
	/** flag wether serial number is set */
	private boolean serialNumberValid = false;
	/** serial number, only valid iff serialNumberValid == true */
	private String serialNumber;
	/** flag wether nodeId is set */
	private boolean nodeIdValid = false;
	/** node id, only valid iff nodeIdValid == true */
	private long nodeId;
	/** reboot flag */
	private boolean reboot = true;
	/** list of upload requests */
	private LinkedList<UploadRequest> requests;
	/** invalidate flag */
	private boolean invalidate = false;
	/** listen flag */
	private boolean listen = false;
	/** connect target address */
	private String connect = "";
	/** connect display flag */
	private boolean connectDisplay = false;
	/** force upload flag */
	private boolean forceUpload = false;
	/** the request which should be executed in memory or null if there is
	 * none */
	private UploadRequest executeInMemory = null;
	/** compression flag */
	private boolean compress = true;
	/** list imotes flag */
	private boolean listImotes = false;

	/** setup a new Configuration */
	public Configuration() {
		this.requests = new LinkedList<UploadRequest>();
	}

	/** parse vendor id
	 *
	 * @param vendorID vendor id
	 */
	public void setVendorId(String vendorID) {
		int id = Integer.parseInt(cutRadix(vendorID), getRadix(vendorID));
		if (id > 0xffff) {
			/* 16 bit unsigned integer */
			throw new NumberFormatException("VendorID > 0xffff");
		}
		this.vendorId = id;
	}

	/** getter for vendor id
	 *
	 * @return vendor id
	 */
	public int getVendorId() {
		return this.vendorId;
	}

	/** write the vendor id to a stream such that it can be parsed by
	 * setVendorId again
	 *
	 * @param level indentaion level
	 * @param writer stream to be written to
	 */
	void printVendorId(int level, PrintStream writer) {
		writer.printf("0x%04x", this.vendorId);
	}

	/** parse product id
	 *
	 * @param productID product id
	 */
	public void setProductId(String productID) {
		int id = Integer.parseInt(cutRadix(productID), getRadix(productID));
		if (id > 0xffff) {
			/* 16 bit unsigned integer */
			throw new NumberFormatException("ProductID > 0xffff");
		}
		this.productId = id;
	}

	/** getter for product id
	 *
	 * @return product id
	 */
	public int getProductId() {
		return this.productId;
	}

	/** write the product id to a stream such that it can be parsed by
	 * setProductId again
	 *
	 * @param level indentaion level
	 * @param writer stream to be written to
	 */
	void printProductId(int level, PrintStream writer) {
		writer.printf("0x%04x", this.productId);
	}

	/** parse serial number
	 *
	 * @param serialNumber serial number
	 */
	public void setSerialNumber(String serialNumber) {
		if (serialNumber.equals("*")) {
			this.serialNumberValid = false;
		} else {
			this.serialNumberValid = true;
			this.serialNumber = serialNumber;
		}
	}

	/** getter for serial number
	 *
	 * @return serial number
	 */
	public String getSerialNumber() {
		return this.serialNumber;
	}

	/** write the serial number to a stream such that it can be parsed by
	 * setSerialNumber again
	 *
	 * @param level indentaion level
	 * @param writer stream to be written to
	 */
	void printSerialNumber(int level, PrintStream writer) {
		if (this.serialNumberValid) {
			writer.print(this.serialNumber);
		} else {
			writer.print("*");
		}
	}

	/** wether serial number is set
	 *
	 * @return true iff serial number is set
	 */
	public boolean hasSerialNumber() {
		return this.serialNumberValid;
	}

	/** parse node id
	 *
	 * @param nodeId node id
	 */
	public void setNodeId(String nodeId) {
		if (nodeId.equals("*")) {
			this.nodeIdValid = false;
		} else {
			this.nodeIdValid = true;
			this.nodeId = Long.parseLong(nodeId);
		}
	}

	/** getter for node id
	 *
	 * @return node id
	 */
	public long getNodeId() {
		return this.nodeId;
	}

	/** write the node id to a stream such that it can be parsed by
	 * setNodeId again
	 *
	 * @param level indentaion level
	 * @param writer stream to be written to
	 */
	void printNodeId(int level, PrintStream writer) {
		if (this.nodeIdValid) {
			writer.print(this.nodeId);
		} else {
			writer.print("*");
		}
	}

	/** wether node id is set
	 *
	 * @return true iff node id is set
	 */
	public boolean hasNodeId() {
		return this.nodeIdValid;
	}

	/** getter for reboot flag
	 *
	 * @return reboot flag
	 */
	public boolean getReboot() {
		return this.reboot;
	}

	/** parse reboot flag
	 *
	 * @param reboot reboot flag
	 */
	public void setReboot(String reboot) {
		this.reboot = Boolean.parseBoolean(reboot);
	}

	/** write the reboot flag to a stream such that it can be parsed by
	 * setReboot again
	 *
	 * @param level indentaion level
	 * @param writer stream to be written to
	 */
	void printReboot(int level, PrintStream writer) {
		writer.print(new Boolean(this.reboot).toString());
	}

	/** add an upload request
	 *
	 * @param req upload request
	 */
	public void addUploadRequest(UploadRequest req) {
		this.requests.add(req);
	}

	/** get upload requests
	 *
	 * @return upload requests
	 */
	public List<UploadRequest> getRequests() {
		return this.requests;
	}

	/** write all upload requests
	 *
	 * @param level indentaion level
	 * @param writer stream to be written to
	 * @throws Exception whenever writing fails
	 */
	void printRequests(int level, PrintStream writer) throws Exception {
		writeClassList(this.requests, level, writer);
	}

	/** parse requests from a stream
	 *
	 * @param reader stream to be read from
	 * @throws Exception whenever reading fails
	 */
	void readRequests(BufferedReader reader) throws Exception {
		this.requests.addAll(readClassList(UploadRequest.class, reader));
	}

	/** getter for invalidate flag
	 *
	 * @return invalidate flag
	 */
	public boolean getInvalidate() {
		return this.invalidate;
	}

	/** parse invalidate flag
	 *
	 * @param invalidate invalidate flag
	 */
	public void setInvalidate(String invalidate) {
		this.invalidate = Boolean.parseBoolean(invalidate);
	}

	/** write the invalidate flag to a stream such that it can be parsed by
	 * setInvalidate again
	 *
	 * @param level indentaion level
	 * @param writer stream to be written to
	 */
	void printInvalidate(int level, PrintStream writer) {
		writer.print(new Boolean(this.invalidate).toString());
	}

	/** getter for force upload flag
	 *
	 * @return force upload flag
	 */
	public boolean getForceUpload() {
		return this.forceUpload;
	}

	/** parse force upload flag
	 *
	 * @param force force upload flag
	 */
	public void setForceUpload(String force) {
		this.forceUpload = Boolean.parseBoolean(force);
	}

	/** write the force upload flag to a stream such that it can be parsed by
	 * setForceUpload again
	 *
	 * @param level indentaion level
	 * @param writer stream to be written to
	 */
	void printForceUpload(int level, PrintStream writer) {
		writer.print(new Boolean(this.forceUpload).toString());
	}

	/** getter for compression flag
	 *
	 * @return compression flag
	 */
	public boolean getCompress() {
		return this.compress;
	}

	/** parse compression flag
	 *
	 * @param compression compression flag
	 */
	public void setCompress(String compress) {
		this.compress = Boolean.parseBoolean(compress);
	}

	/** write the compression flag to a stream such that it can be parsed by
	 * setCompress again
	 *
	 * @param level indentaion level
	 * @param writer stream to be written to
	 */
	void printCompress(int level, PrintStream writer) {
		writer.print(new Boolean(this.compress).toString());
	}

	/** getter for listen flag
	 *
	 * @return listen flag
	 */
	public boolean getListen() {
		return this.listen;
	}

	/** getter for list imotes flag
	 *
	 * @return list imotes flag
	 */
	public boolean getListImotes() {
		return this.listImotes;
	}

	/** parse connect string
	 *
	 * @param connect connect string
	 */
	public void setConnect(String connect) {
		if (connect.equals("*")) {
			this.connectDisplay = true;
			connect = System.getenv("DISPLAY").replaceAll(":.*", "");
		} else {
			this.connectDisplay = false;
		}
		this.connect = connect;
	}

	/** getter for connect string
	 *
	 * @return connect string
	 */
	public String getConnect() {
		return this.connect;
	}

	/** write the connect string to a stream such that it can be parsed by
	 * setInvalidate again
	 *
	 * @param level indentaion level
	 * @param writer stream to be written to
	 */
	void printConnect(int level, PrintStream writer) {
		if (this.connectDisplay) {
			writer.print("*");
		} else {
			writer.print(this.connect);
		}
	}

	/** wether connect string is set
	 * 
	 * @return true iff connect option is set
	 */
	public boolean hasConnect() {
		return !this.connect.equals("");
	}

	/** helper for writing a collection of a specific class
	 *
	 * @param <tp> component type of collection
	 * @param list list to be written out
	 * @param level indentation level
	 * @param writer stream to be written to
	 * @throws Exception whenever writing goes wrong
	 */
	private static <tp> void writeClassList(Collection<tp> list, int level,
		PrintStream writer) throws Exception {
		writer.println("{");
		for (tp elem : list) {
			String tabs = buildTabs(level + 1);
			writer.println(tabs + "{");
			/* write out element */
			writeClass(elem, level + 2, writer);
			writer.println(tabs + "}");
		}
		writer.print(buildTabs(level) + "}");
	}

	/** helper for reading a collection of a specific class
	 *
	 * @param <tp> component type of collection
	 * @param elemClass type
	 * @param reader stream to be read from
	 * @return collection of instances of the class
	 * @throws Exception whenever anything goes wrong
	 */
	private static <tp> LinkedList<tp> readClassList(Class<tp> elemClass,
		BufferedReader reader) throws Exception {
		LinkedList<tp> list = new LinkedList<tp>();
		String line;

		while ((line = reader.readLine()) != null) {
			/* simple parsing */
			line = line.replaceFirst("^\\s+", "");
			if (line.matches("}\\s*$")) {
				return list;
			}
			if (!line.matches("\\{\\s*$")) {
				throw new Error();
			}
			tp request = (tp) elemClass.newInstance();
			/* read element and add */
			readClass(request, reader);
			list.add(request);
		}

		throw new Error();
	}

	/** helper for writing out an instance of a class
	 *
	 * @param <tp> type of class
	 * @param instance instance of class
	 * @param level indentation level
	 * @param writer stream to be written to
	 * @throws Exception whenever writing goes wrong
	 */
	private static <tp> void writeClass(tp instance, int level,
		PrintStream writer) throws Exception {
		for (Method m : instance.getClass().getDeclaredMethods()) {
			String name = m.getName();
			/* consider all "print..." methods */
			if (name.startsWith("print")) {
				String key = name.substring(5);
				writer.print(buildTabs(level) + key + " = ");
				/* invoke print method to write value */
				m.invoke(instance, level, writer);
				writer.print("\n");
			}
		}
	}

	/** helper for reading an instance of a specific class
	 *
	 * @param <tp> type
	 * @param instance instance to be read into
	 * @param reader stream to be read from
	 * @throws Exception whenever anything goes wrong
	 */
	private static <tp> void readClass(tp instance, BufferedReader reader)
		throws Exception {
		String line;

		while ((line = reader.readLine()) != null) {
			/* simple parsing */
			line = line.replaceFirst("^\\s+", "");
			if (line.matches("}\\s*$")) {
				return;
			}
			int pos = line.indexOf("=");
			if (pos == -1) {
				continue;
			}
			String key = line.substring(0, pos).replaceAll("\\s+$", "");
			String value = line.substring(pos + 1).replaceFirst("^\\s+", "");
			try {
				/* try setter method first if there is one */
				Method m = instance.getClass().getDeclaredMethod("set" + key,
					String.class);
				m.invoke(instance, value);
			} catch (NoSuchMethodException e) {
				/* try read method next */
				Method m = instance.getClass().getDeclaredMethod("read" + key,
					BufferedReader.class);
				m.invoke(instance, reader);
			}
		}
	}

	/** debug method to dump config on stdout */
	public void dump() {
		System.out.println("VendorId: " + String.format("0x%04x", this.vendorId));
		System.out.println("ProductId: " + String.format("0x%04x",
			this.productId));
		System.out.println("Serial Number: " + (this.serialNumberValid
			? String.format("0x%s", this.serialNumber) : "<any>"));
		System.out.println("Uploads Requests:");
		for (UploadRequest request : this.requests) {
			System.out.println("\tFile: " + request.getFileName()
				+ " Target: " + String.format("0x%08x",
				request.getTargetAddress())
				+ " Execute in RAM: " + (request.getExecuteRam() ? "yes" : "no"));
		}
		System.out.println("Invalidate default image Location: "
			+ new Boolean(this.invalidate).toString());
		System.out.println("Force Uploads: "
			+ new Boolean(this.forceUpload).toString());
		System.out.println("Compress Uploads: "
			+ new Boolean(this.compress).toString());
		System.out.println("Role: "
			+ (this.listen ? "Server"
			: this.hasConnect() ? "Client, connecting to " + this.connect
			: "Standalone"));
	}

	/** helper function to generate tabs for indentation
	 *
	 * @param t indentation level
	 * @return t tabs
	 */
	private static String buildTabs(int t) {
		StringBuilder ret = new StringBuilder(t);

		for (int i = 0; i < t; ++i) {
			ret.append('\t');
		}

		return ret.toString();
	}

	/** helper function to parse an integer
	 *
	 * @param str integer as string
	 * @return integer as string without prefix
	 */
	static private String cutRadix(String str) {
		int radix = getRadix(str);
		if (radix == 16) {
			return str.substring(2);
		} else if (radix == 8) {
			return str.substring(1);
		} else {
			return str;
		}
	}

	/** helper function to get radix of an string encoded integer
	 *
	 * @param str integer as string
	 * @return radix
	 */
	static private int getRadix(String str) {
		if (str.length() >= 2 && str.substring(0, 2).equals("0x")) {
			return 16;
		} else if (str.length() > 1 && str.substring(0, 1).equals("0")) {
			return 8;
		} else {
			return 10;
		}
	}

	/** parse command line
	 *
	 * @param args arguments as given to main
	 * @throws Exception whenever anything goes wrong
	 */
	public void parseCommandLine(String[] args) throws Exception {
		/* declare options */
		Option vidOption = OptionBuilder.withArgName("vendor-id").hasArg().
			withDescription("Vendor ID of the Imote. Defaults to "
			+ String.format("0x%04x", Configuration.DEFAULT_VENDOR_ID)).create(
			"vid");

		Option pidOption = OptionBuilder.withArgName("product-id").hasArg().
			withDescription("Product ID of the Imote. Defaults to "
			+ String.format("0x%04x", Configuration.DEFAULT_PRODUCT_ID)).create(
			"pid");

		Option snrOption = OptionBuilder.withArgName("serialnumber").hasArg().
			withDescription("Serial Number of the Imote. Type '*' for any.").
			create("snr");

		Option nidOption = OptionBuilder.withArgName("nodeid").hasArg().
			withDescription("Node ID of the Imote. Type '*' for any.").
			create("nid");

		Option fileOption = OptionBuilder.withArgName("file:target").hasArgs(1).
			hasOptionalArgs(2).withDescription("Adds a file to upload with its "
			+ "optional target address").withValueSeparator(':').create("file");

		Option execRamOption = OptionBuilder.withArgName("execram").
			withDescription("Loads the image into the ram and executes it there").
			create("execram");

		Option writeOption = OptionBuilder.withArgName("write").hasArg().
			withDescription("writes a new config file").create("write");

		Option dryrunOption = OptionBuilder.withArgName("dryrun").
			withDescription("Reads and Writes Config File(s) and exits then").
			create("dryrun");

		Option rebootOption = OptionBuilder.withArgName("noreboot").
			withDescription("Does not reboot the device afterwards").create(
			"noreboot");

		Option invalidateOption = OptionBuilder.withArgName("invalidate").
			withDescription("Invalidate default image Location").create(
			"invalidate");

		Option forceUploadOption = OptionBuilder.withArgName("forceUpload").
			withDescription("Upload image even if it does not pass the"
			+ "validity check").create("forceUpload");

		Option noCompressOption = OptionBuilder.withArgName("nocompress").
			withDescription("Do not compress uploads").create("nocompress");

		Option listenOption = OptionBuilder.withArgName("listen").
			withDescription("Run in Listening mode").create("listen");

		Option listImotesOption = OptionBuilder.withArgName("listImotes").
			withDescription("List connected Imotes").create("listImotes");

		Option connectOption = OptionBuilder.withArgName("connect").hasArg().
			withDescription(
			"Connect to a daemon specified by its network address").create(
			"connect");

		Option connectDisplayOption = OptionBuilder.withArgName("connectdisplay").
			withDescription(
			"Connect to a daemon specified by the DISPLAY environment variable").
			create("connectdisplay");

		Option helpOption = new Option("help", "print this help message");

		/* add options */
		Options options = new Options();
		options.addOption(helpOption);
		options.addOption(vidOption);
		options.addOption(pidOption);
		options.addOption(snrOption);
		options.addOption(nidOption);
		options.addOption(fileOption);
		options.addOption(execRamOption);
		options.addOption(writeOption);
		options.addOption(dryrunOption);
		options.addOption(rebootOption);
		options.addOption(invalidateOption);
		options.addOption(forceUploadOption);
		options.addOption(noCompressOption);
		options.addOption(listenOption);
		options.addOption(listImotesOption);
		options.addOption(connectOption);
		options.addOption(connectDisplayOption);

		CommandLineParser parser = new GnuParser();

		try {
			/* parse cmd line */
			CommandLine cmd = parser.parse(options, args);

			/* evaluate parsed options */

			if (cmd.hasOption("help")) {
				HelpFormatter formatter = new HelpFormatter();
				formatter.printHelp("codeloader", options, true);
				System.exit(0);
			}

			if (cmd.hasOption("listen")) {
				this.listen = true;
				return;
			}

			String[] configFiles = cmd.getArgs();
			for (String configFile : configFiles) {
				if (configFile.equals("-")) {
					this.readStdin();
				} else {
					this.readFile(configFile);
				}
			}

			if (cmd.hasOption("vid")) {
				String val = cmd.getOptionValue("vid");
				this.setVendorId(val);
			}

			if (cmd.hasOption("pid")) {
				String val = cmd.getOptionValue("pid");
				this.setProductId(val);
			}

			if (cmd.hasOption("snr")) {
				String val = cmd.getOptionValue("snr");
				this.setSerialNumber(val);
			}

			if (cmd.hasOption("nid")) {
				String val = cmd.getOptionValue("nid");
				this.setNodeId(val);
			}

			if (cmd.hasOption("file")) {
				String[] val = cmd.getOptionValues("file");
				Configuration.UploadRequest request =
					new Configuration.UploadRequest();
				request.setFileName(val[0]);
				if (val.length == 2) {
					request.setTargetAddress(val[1]);
				}
				if (cmd.hasOption("execram")) {
					request.setExecuteRam(true);
				}
				this.addUploadRequest(request);
			}

			if (cmd.hasOption("noreboot")) {
				this.reboot = false;
			}

			if (cmd.hasOption("invalidate")) {
				this.invalidate = true;
			}

			if (cmd.hasOption("forceUpload")) {
				this.forceUpload = true;
			}

			if (cmd.hasOption("nocompress")) {
				this.compress = false;
			}

			if (cmd.hasOption("listImotes")) {
				this.listImotes = true;
			}

			if (cmd.hasOption("connect")) {
				String val = cmd.getOptionValue("connect");
				this.setConnect(val);
			}

			if (cmd.hasOption("connectdisplay")) {
				this.setConnect("*");
			}

			if (!this.checkRamExecution()) {
				System.err.println("Cannot use more than one image for "
					+ "RAM execution");
				System.exit(1);
			}

			if (cmd.hasOption("write")) {
				String val = cmd.getOptionValue("write");
				if (val.equals("-")) {
					this.writeStdout();
				} else {
					this.writeFile(val);
				}
			}

			if (cmd.hasOption("dryrun")) {
				System.exit(0);
			}
		} catch (ParseException e) {
			System.err.println("Parsing options failed: " + e.getMessage());
			HelpFormatter formatter = new HelpFormatter();
			formatter.printHelp("codeloader", options, true);
			System.exit(1);
		}
	}

	/** iterate over all upload requests and check wether some sideconditions
	 * are met
	 *
	 * @return true if all side conditions are met
	 */
	private boolean checkRamExecution() {
		for (UploadRequest req : this.requests) {
			if (req.getExecuteRam() == true) {
				if (this.executeInMemory != null) {
					return false;
				}
				this.executeInMemory = req;
			}
		}
		return true;
	}

	/** wether there is an execute-in-ram request
	 *
	 * @return true iff there is an execute-in-ram request
	 * @note invoke checkRamExecution before that
	 */
	public boolean hasRamExecutionRequest() {
		return this.executeInMemory != null;
	}

	/** get execute-in-ram request
	 *
	 * @return execute-in-ram request or null if there is none
	 * @note invoke checkRamExecution before that
	 */
	public UploadRequest getRamExecutionRequest() {
		return this.executeInMemory;
	}

	/** read a configuration file
	 *
	 * @param fileName path to file
	 * @throws Exception whenever anything goes wrong
	 */
	public void readFile(String fileName) throws Exception {
		FileInputStream reader = new FileInputStream(fileName);
		readClass(this, new BufferedReader(new InputStreamReader(reader)));
	}

	/** read a configuration from stdin
	 *
	 * @throws Exception whenever anything goes wrong
	 */
	public void readStdin() throws Exception {
		readClass(this, new BufferedReader(new InputStreamReader(System.in)));
	}

	/** write a configuration file
	 *
	 * @param fileName path to file
	 * @throws Exception whenever anything goes wrong
	 */
	public void writeFile(String fileName) throws Exception {
		PrintStream outFile = new PrintStream(fileName);
		writeClass(this, 0, outFile);
	}

	/** write a configuration to stdout
	 *
	 * @throws Exception whenever anything goes wrong
	 */
	public void writeStdout() throws Exception {
		writeClass(this, 0, System.out);
	}
}
